import paramiko
import logging
import time
import json
import os
import string

from random import choice
from typing import Tuple, Optional


def makeTmpFile(sshClient: paramiko.SSHClient) -> bytes:
    _, stdoutReader, stderrReader = sshClient.exec_command("mktemp")
    stderr: bytes = stderrReader.read()
    if stderr:
        logging.warning("Got %s in stderr, exiting..." % stderr)
        exit(1)  # todo: reinsert into redis or something?
    stdout: bytes = stdoutReader.read()
    logging.info("Got temp file: %s" % stdout.strip())
    return stdout.strip()


def paramikoExecuteRead(sshClient: paramiko.SSHClient, cmdStr: str, skipStdout: bool = False, _sync: bool = True) -> Tuple[bytes, bytes]:
    # We can't rely on fabric or paramiko (for god knows what reason), so we run our command in the background,
    # monitor the pid, and then read the output from temp files once we're done
    stderrFile: bytes = makeTmpFile(sshClient)
    stdoutFile: bytes = makeTmpFile(sshClient)

    screenName = ''.join([choice(string.ascii_lowercase) for _ in range(10)])

    if skipStdout:
        screenCmd = "screen -dm -S %s bash -c '%s 2> %s'; screen -S %s -Q echo '$PID'" % \
                    (screenName, cmdStr, stderrFile.decode('utf-8'), screenName)
    else:
        screenCmd = "screen -dm -S %s bash -c '%s > %s 2> %s'; screen -S %s -Q echo '$PID'" % \
                    (screenName, cmdStr, stdoutFile.decode('utf-8'), stderrFile.decode('utf-8'), screenName)

    _, stdout, stderr = sshClient.exec_command(screenCmd, timeout=10)

    pid = stdout.readline()
    if not pid:
        raise Exception("No pid found!!! cmd: %s" % screenCmd)

    while pidIsAlive(sshClient, pid):
        time.sleep(1)

    # Sleep one last time to make sure the files have been flushed
    time.sleep(10)
    if _sync:
        paramikoExecuteRead(sshClient, "sync", skipStdout=True, _sync=False)

    return readRemoteFile(sshClient, stdoutFile), readRemoteFile(sshClient, stderrFile)


def pidIsAlive(sshClient: paramiko.SSHClient, pid: str) -> bool:
    chan = sshClient.get_transport().open_session()
    chan.exec_command("kill -0 %s" % pid)
    return chan.recv_exit_status() == 0


def readRemoteFile(sshClient: paramiko.SSHClient, filename: bytes) -> bytes:
    sftpClient: paramiko.SFTP = sshClient.open_sftp()
    with sftpClient.open(filename.decode("utf-8"), "r") as f:
        return f.read()

def getSSHConnection() -> paramiko.SSHClient:
    hostList, username = os.environ["SCANNING_SERVER_IPS"], os.environ["SCANNING_SERVER_USER"]
    host = choice(json.loads(hostList))
    sshClient = paramiko.SSHClient()
    sshClient.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        sshClient.connect(hostname=host, username=username,
                          password=os.environ["SCANNING_SERVER_PASSWORD"], port=2222)
        logging.info("Connected to host=%s" % host)
        return sshClient
    except paramiko.ssh_exception.SSHException as e:
        logging.info("Failed to connect to host=%s" % host)
        raise e

def writeDataToTmpFile(sshClient: paramiko.SSHClient, data: str) -> str:
    filename: bytes = makeTmpFile(sshClient)
    sftpClient: paramiko.SFTP = sshClient.open_sftp()
    with sftpClient.open(filename.decode('utf-8'), 'w') as f:
        f.write(data)
    return filename.decode("utf-8")


from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, ForeignKey, Sequence, Boolean, Text, DateTime, LargeBinary

from tulgey.db import Base
from tulgey.PortScanResult import PortScanResult, PortStatus
from tulgey.IPScanResult import IPScanResult
from tulgey.utils.iputils import normalizeIPv6

class PortScanResultModel(Base):
    __tablename__ = "portscanresult"
    id = Column(Integer, primary_key=True)
    port = Column(Integer, nullable=False)
    status = Column(Text, nullable=False)  # "OPEN" || "CLOSED"
    serviceName = Column(Text, nullable=True)
    banner = Column(Text, nullable=True)

    ipScanResultModel_id = Column(Integer, ForeignKey("ipscanresult.id"))

    @staticmethod
    def frum(portScanResult):
        return PortScanResultModel(
            port=portScanResult.port,
            status=str(portScanResult.status),
            serviceName=portScanResult.serviceName,
            banner=portScanResult.banner,
        )

    def to(self) -> PortScanResult:
        return PortScanResult(
            port=self.port,
            status=(PortStatus.fromString(self.status)),
            serviceName=self.serviceName,
            banner=self.banner,
        )

class IPScanResultModel(Base):
    __tablename__ = "ipscanresult"
    id = Column(Integer, primary_key=True)
    isUp = Column(Boolean, nullable=False)
    scannedPorts = Column(Text, nullable=False)
    datetime = Column(DateTime, nullable=False)

    ip_id = Column(Text, ForeignKey("ip.ip"), index=True)
    portScanResults = relationship("PortScanResultModel", backref="ipscanresult")

    @staticmethod
    def frum(ipScanResult: IPScanResult):
        return IPScanResultModel(
            ip_id=normalizeIPv6(ipScanResult.ip),
            isUp=ipScanResult.isUp,
            scannedPorts=ipScanResult.scannedPorts,
            portScanResults=[
                PortScanResultModel.frum(psr) for psr in ipScanResult.ports.values()
            ],
            datetime=ipScanResult.dt,
        )

    def to(self) -> IPScanResult:
        return IPScanResult(
            ip=normalizeIPv6(self.ip_id),
            isUp=self.isUp,
            ports={psrm.to().port:psrm.to() for psrm in self.portScanResults},
            scannedPorts=self.scannedPorts,
            dt=self.datetime,
        )


class IP(Base):
    __tablename__ = "ip"
    ip = Column(Text, primary_key=True, nullable=False)
    lastScanDate = Column(DateTime, nullable=True, index=True)  # null if it has never been scanned
    lastTraceDate = Column(DateTime, nullable=True, index=True)  # null if it has never been traced
    sources = Column(Text, nullable=False)
    isIPv4 = Column(Boolean, nullable=False)
    everPingable = Column(Boolean, nullable=True)

    ipScanResults = relationship("IPScanResultModel", backref="ip")


import socket
from typing import List
import ipaddress

def isValidIPv6(address: str) -> bool:
    try:
        socket.inet_pton(socket.AF_INET6, address)
        return True
    except socket.error:
        return False

def normalizeIPv6(address: str) -> str:
    return socket.inet_ntop(socket.AF_INET6, socket.inet_pton(socket.AF_INET6, address.strip()))

def isValidIPv4(address: str) -> bool:
    try:
        socket.inet_pton(socket.AF_INET, address)
        return True
    except socket.error:
        return False

def isValidIPv4Range(ipRange: str) -> bool:
    return ipRange.count("/") == 1 and isValidIPv4(ipRange.split("/")[0])

def expandIPv4Range(ipRange: str) -> List[str]:
    net = ipaddress.ip_network(ipRange)
    ret = []
    for host in net.hosts():
        ret.append(host.exploded)
    return ret


from typing import Mapping, Any
from enum import Enum
from dataclasses import dataclass


class PortStatus(Enum):
    OPEN = "OPEN"
    CLOSED = "CLOSED"
    # Are there other ones? Like filtered? Do we care? Does masscan care?

    def __repr__(self):
        return "PortStatus." + self.name

    def __str__(self):
        return self.name

    @staticmethod
    def fromString(str: str) -> "PortStatus":
        if str == "OPEN":
            return PortStatus.OPEN
        if str == "CLOSED":
            return PortStatus.CLOSED
        raise Exception("Got invalid PortStatus=%s" % str)


@dataclass(frozen=True)
class PortScanResult:
    port: int
    status: PortStatus
    serviceName: str
    banner: str

    def toJsonDict(self) -> Mapping[str, Any]:
        return {
            "port": self.port,
            "status": str(self.status),
            "serviceName": self.serviceName,
            "banner": self.banner,
        }

    @staticmethod
    def fromJsonDict(jsonDict) -> "PortScanResult":
        return PortScanResult(
            jsonDict["port"],
            PortStatus.fromString(jsonDict["status"]),
            jsonDict["serviceName"],
            jsonDict["banner"],
        )



from datetime import datetime

DATETIME_FORMAT = "%Y %m %d %H %M %S"


def toString(datetime: datetime) -> str:
    return datetime.strftime(DATETIME_FORMAT)


def fromString(dateStr: str) -> datetime:
    return datetime.strptime(dateStr, DATETIME_FORMAT)
